/* helpfile.c - Help file loading and display
 * Copyright 2000-2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: helpfile.c,v 1.34.2.13 2001/11/18 00:44:48 entrope Exp $
 */

#include "config.h"
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "common.h"
#include "dict.h"
#include "hash.h"
#include "nickserv.h"
#include "proto.h"
#include "recdb.h"

#include "helpfile.h"

#define HFMSG_HELP_NOT_STRING	"Help file error (help data was not a string)."

struct helpfile {
    dict_t db;

    expand_func_t expand;
};

extern struct userNode *global, *chanserv, *opserv, *nickserv;

static int
vsend_message(const char *dest, struct userNode *src, struct handle_info *handle, int use_privmsg, expand_func_t expand, const char *format, va_list al)
{
    void (*irc_send)(const char *from, const char *to, const char *msg);
    char *input, line[MAX_LINE_SIZE];
    unsigned int size, pos, length, chars_sent;
    int ret, use_color;

    /* expand the parameters we got .. */
    chars_sent = 0;
    /* Normally, I'd pass NULL as the first parameter to vsnprintf.
     * But guess what?  It makes FreeBSD's libc allocate a buffer that it
     * never frees!  Surprise!
     * The kludge-around is to pass a non-NULL pointer instead.
     */
    if ((ret = vsnprintf(line, 0, format, al)) <= 0) {
        char *temp;
	/* non-C99 behavior */
	size = strlen(format);
	input = malloc(size);
	while ((ret = vsnprintf(input, size, format, al)) == -1) {
	    size <<= 1;
	    input = realloc(input, size);
	}
        temp = alloca(ret+1);
        strcpy(temp, input);
        free(input);
        input = temp;
    } else {
	/* C99 is nice about how vsnprintf fails */
	size = ret + 1;
	input = alloca(size);
	ret = vsnprintf(input, size, format, al);
    }

    /* figure out how to send the messages */
    if (handle) {
	use_privmsg |= HANDLE_FLAGGED(handle, USE_PRIVMSG);
	use_color = HANDLE_FLAGGED(handle, MIRC_COLOR);

	if (handle->screen_width) {
	    /* NickServ should clamp this value anyway. */
	    size = handle->screen_width;
	}
        if (size > sizeof(line)) size = sizeof(line);
    } else {
        size = sizeof(line);
        use_color = 1;
    }
    irc_send = use_privmsg ? irc_privmsg : irc_notice;

    /* This used to be two passes, but if you do that and allow
     * arbitrary sizes for ${}-expansions (as with help indexes),
     * that requires a very big intermediate buffer.
     */
    pos = 0;
    while (*input) {
	char *value, *free_value;

	while (*input && *input != '$' && *input != '\n' && pos < size) {
	    line[pos++] = *input++;
	}

	if (!*input) goto send_line;
        if (*input == '\n') {
            input++;
            goto send_line;
        }
	if (pos == size) {
            char *terminal;
            line[size - 1] = 0;
            terminal = strrchr(line, ' ');
            if (terminal) {
                /* This is goofy.  Having -1 screws it up if colorizing is on,
                 * having -0 screws it up if colorizing is off.  There must be
                 * some other glitch somewhere in here conspiring with this.
                 */
                int backup = pos - (terminal - line) - 1;
                *terminal = 0;
                input -= backup;
                pos -= backup;
            } else {
                /* Send the entire word, even though it's bigger than
                 * the requested maximum size.  But stop at line size.
                 */
                while (*input && (*input != ' ') && (pos < MAX_LINE_SIZE)) line[pos++] = *input++;
            }
            while (*input == ' ') input++;
	    goto send_line;
	}

        free_value = 0;
	switch (*(++input)) {
        /* Literal '$' or end of string. */
	case 0:
	    input--;
	case '$':
	    value = "$";
	    break;
	/* The following two expand to mIRC color codes if enabled
	   by the user. */
	case 'b':
	    value = use_color ? "\002" : "";
	    break;
	case 'o':
	    value = use_color ? "\017" : "";
	    break;
	case 'u':
	    value = use_color ? "\037" : "";
	    break;
	/* Service nicks. */
	case 'G':
	    value = global ? global->nick : "Global";
	    break;
	case 'C':
	    value = chanserv ? chanserv->nick : "ChanServ";
	    break;
	case 'O':
	    value = opserv ? opserv->nick : "OpServ";
	    break;
	case 'N':
            value = nickserv ? nickserv->nick : "NickServ";
            break;
        case 'n':
            if (nickserv) {
                length = strlen(nickserv->nick);
                value = alloca(length+strlen(nickserv->hostname)+2);
                strcpy(value, nickserv->nick);
                value[length++] = '@';
                strcpy(value+length, nickserv->hostname);
            } else {
                value = "NickServ@Services";
            }
	    break;
	case 'H':
	    value = handle ? handle->handle : "Handle";
	    break;
	/* Custom expansion handled by helpfile-specific function. */
	case '{':
	    if (expand) {
		char *name_end = input + 1;

		while (*name_end != '}' && *name_end) name_end++;
		if (*name_end) {
		    *name_end = 0;
		    value = free_value = expand(input + 1);
                    if (!value) value = "";
		    input = name_end;
		    break;
		}
	    }

	/* Let it fall through when there's no expansion function or
	terminating '}'. */
	default:
		value = alloca(3);
		value[0] = '$';
		value[1] = *input;
		value[2] = 0;
	}
	input++;
#define SEND_LINE() do { line[pos] = 0; irc_send(src->numeric, dest, line); chars_sent += pos; pos = 0; } while (0)
        while ((pos + strlen(value) > size) || strchr(value, '\n')) {
            unsigned int avail;
            avail = size - pos - 1;
            length = strcspn(value, "\n ");
            if (length <= avail) {
                strncpy(line+pos, value, length);
                pos += length;
                value += length;
                /* copy over spaces, until (possible) end of line */
                while (*value == ' ') {
                    if (pos < size-1) line[pos++] = *value;
                    value++;
                }
            } else {
                /* word to send is too big to send now.. what to do? */
                if (pos > 0) {
                    /* try to put it on a separate line */
                    SEND_LINE();
                } else {
                    /* already at start of line; only send part of it */
                    strncpy(line, value, avail);
                    pos += avail;
                    value += length;
                    /* skip any trailing spaces */
                    while (*value == ' ') value++;
                }
            }
            /* if we're looking at a newline, send the accumulated text */
            if (*value == '\n') {
                SEND_LINE();
                value++;
            }
        }
        length = strlen(value);
	memcpy(line + pos, value, length);
        if (free_value) free(free_value);
	pos += length;
        if ((pos < size-1) && *input) continue;
      send_line:
        SEND_LINE();
#undef SEND_LINE
    }
    return chars_sent;
}

int
send_message(struct userNode *dest, struct userNode *src, const char *format, ...)
{
    int res;
    va_list ap;

    va_start(ap, format);
    res = vsend_message(dest->numeric, src, dest->handle_info, 0, NULL, format, ap);
    va_end(ap);
    return res;
}

int
send_target_message(int use_privmsg, const char *dest, struct userNode *src, const char *format, ...)
{
    int res;
    va_list ap;

    va_start(ap, format);
    res = vsend_message(dest, src, NULL, use_privmsg, NULL, format, ap);
    va_end(ap);
    return res;
}

void column_send(struct userNode *service, struct userNode *user, struct string_list *list)
{
    unsigned int n, column, len, max, pos, columns, size;
    char buffer[MAX_LINE_SIZE];

    for(n = max = len = 0; n < list->used; n++)
    {	
	if((len = strlen(list->list[n])) > max) max = len;
    }

    if(!max)
    {
	send_message(user, service, "None.");
	return;
    }

    size = user->handle_info ? user->handle_info->screen_width : max * 4;

    /* Since columns look best when no wrapping occurs at the client end,
       restrict the line size a bit. */
    if(size > COLUMN_LINE_SIZE)
    {
	size = COLUMN_LINE_SIZE;
    }

    columns = size / max;

    if(!columns)
    {
	columns = 1;
    }

    for (n = column = pos = 0; n < list->used; n++)
    {
	if(column++ >= columns)
	{
	    send_message(user, service, "%s", buffer);
	    pos = 0;
	    column = 1;
	}

	if(column > 1)
	{
	    (void)memset(buffer + pos, ' ', (max - len) + 3);
	    pos += (max - len) + 3;
	}

	pos += (len = sprintf(buffer + pos, "%s", list->list[n]));
	buffer[pos] = 0;
    }

    if(column)
    {
	send_message(user, service, "%s", buffer);
    }
}

int
_send_help(struct userNode *dest, struct userNode *src, expand_func_t expand, const char *format, ...)
{
    int res;

    va_list ap;
    va_start(ap, format);
    res = vsend_message(dest->numeric, src, dest->handle_info, 0, expand, format, ap);
    va_end(ap);
    return res;
}

int
send_help(struct userNode *dest, struct userNode *src, struct helpfile *hf, const char *topic)
{
    struct record_data *rec;
    const char *str;
    if (!topic) topic = "<index>";
    if (hf) {
	rec = dict_find(hf->db, topic, NULL);
    } else {
	rec = NULL;
    }
    if (!rec) {
	rec = dict_find(hf->db, "<missing>", NULL);
	if (!rec) {
	    _send_help(dest, src, hf->expand, MSG_TOPIC_UNKNOWN);
	    return 0;
	}
    }
    if (rec->type != RECDB_QSTRING) {
	_send_help(dest, src, hf->expand, HFMSG_HELP_NOT_STRING);
	return 0;
    }
    str = GET_RECORD_QSTRING(rec);
    return _send_help(dest, src, hf->expand, str);
}

int
unlistify_help(const char *key, void *data, void *extra)
{
    struct record_data *rd = data;
    dict_t newdb = extra;
    key = strdup(key);
    if (rd->type == RECDB_QSTRING) {
	dict_insert(newdb, key, alloc_record_data_qstring(GET_RECORD_QSTRING(rd)));
	return 0;
    } else if (rd->type == RECDB_STRING_LIST) {
	struct string_list *slist = GET_RECORD_STRING_LIST(rd);
	char *dest;
	unsigned int totlen, len, i;
	for (i=totlen=0; i<slist->used; i++) {
	    totlen = totlen + strlen(slist->list[i]) + 1;
	}
	dest = alloca(totlen+1);
	for (i=totlen=0; i<slist->used; i++) {
	    len = strlen(slist->list[i]);
	    memcpy(dest+totlen, slist->list[i], len);
	    dest[totlen+len] = '\n';
	    totlen = totlen + len + 1;
	}
	dest[totlen] = 0;
	dict_insert(newdb, key, alloc_record_data_qstring(dest));
	return 0;
    } else {
	return 1;
    }
}

struct helpfile *
open_helpfile(const char *fname, expand_func_t expand)
{
    struct helpfile *hf;
    dict_t db = parse_database(fname);
    hf = malloc(sizeof(struct helpfile));
    hf->db = alloc_database();
    hf->expand = expand;
    if (db) {
	dict_foreach(db, unlistify_help, hf->db);
	free_database(db);
    }
    return hf;
}

void close_helpfile(struct helpfile *hf)
{
    free_database(hf->db);
    free(hf);
}
