/* sockcheck.c - insecure proxy checking
 * Copyright 2000-2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: sockcheck.c,v 1.54.2.38 2001/12/05 22:04:59 entrope Exp $
 */

#include "config.h"
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_SYS_POLL_H
#include <sys/poll.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "common.h"
#include "conf.h"
#include "dict.h"
#include "gline.h"
#include "log.h"
#include "recdb.h"
#include "sockcheck.h"

static int disable_sockcheck;

#if SOCKCHECK_USABLE

#if !defined(SOCKCHECK_DEBUG)
#define SOCKCHECK_DEBUG 0
#endif
#define SOCKCHECK_TEST_DB "sockcheck.db"

#define TV_CMP(a, CMP, b) (((a).tv_sec CMP (b).tv_sec) || (((a).tv_sec == (b).tv_sec) && ((a).tv_usec CMP (b).tv_usec)))

/* Pipe to feed worker thread new jobs. */
static int pending_ip_pipe[2];

enum sockcheck_decision {
    PENDING,
    CHECKING,
    ACCEPT,
    REJECT,
};

typedef struct {
    enum sockcheck_decision decision;
    time_t last_touched;
    const char *reason;
    unsigned long addr;
    unsigned char hostname[0];
} *sockcheck_cache_info;

DECLARE_LIST(sci_list, sockcheck_cache_info);
DEFINE_LIST(sci_list, sockcheck_cache_info);

/* And here's the list of hostnames that need to be started on.
 * This is primarily used by the worker thread, although the main
 * thread may read the length to guess how much "behind" the checker
 * is in terms of new clients.
 */
static struct string_list *pending_ip_list;

/* Map of previously checked IPs state (if we've accepted the address yet).
 * Protected by checked_ip_mutex.
 * The data for each entry is a pointer to one of the dec_* variables.
 */
static dict_t checked_ip_dict;
static pthread_mutex_t checked_ip_mutex = PTHREAD_MUTEX_INITIALIZER;

/* Each sockcheck template is formed as a Mealy state machine (that is,
 * the output on a state transition is a function of both the current
 * state and the input).  Mealy state machines require fewer states to
 * match the same input than Moore machines (where the output is only
 * a function of the current state).
 * 
 * A state is characterized by sending some data (possibly nothing),
 * waiting a certain amount of time to receive one of zero or more
 * responses, and a decision (accept, reject, continue to another
 * state) based on the received response.
 */

struct sockcheck_response {
    const char *template;
    struct sockcheck_state *next;
};

DECLARE_LIST(response_list, struct sockcheck_response *);
DEFINE_LIST(response_list, struct sockcheck_response *);
static unsigned int max_responses;

struct sockcheck_state {
    unsigned short port;
    unsigned short timeout;
    unsigned short reps;
    enum sockcheck_decision type;
    const char *template;
    struct response_list responses;
};

struct sockcheck_list {
    unsigned int size, used, refs;
    struct sockcheck_state **list;
};

/*
 * List of tests protected by tests_mutex.
 */
static struct sockcheck_list *tests;
static pthread_mutex_t tests_mutex = PTHREAD_MUTEX_INITIALIZER;

/* Stuff used inside the worker thread, one instance per open connection. */
struct sockcheck_client {
    struct sockcheck_list *tests;
    sockcheck_cache_info addr;
    unsigned int test_index;
    unsigned short test_rep;
    struct sockcheck_state *state;
    unsigned int send_size, send_used, send_pos;
    unsigned char *send;
    unsigned int read_size, read_used, read_pos;
    unsigned char *read;
    const unsigned char **resp_state;
    struct timeval read_timeout;
    unsigned int connected : 1;
    unsigned int writing : 1;
};

static struct {
    unsigned int max_sockets;
    unsigned int max_read;
    struct sockaddr_in *local_addr;
    int local_addr_len;
    time_t max_cache_age;
} sockcheck_conf;

static pthread_t sockcheck_thread;
static unsigned int sockcheck_num_sockets;
static struct pollfd *poll_list;
static struct sockcheck_client **client_list;
static struct timeval tv_now;
static time_t last_clean;
static unsigned int proxies_detected, checked_ip_count;

static struct sockcheck_list *
sockcheck_list_alloc(unsigned int size)
{
    struct sockcheck_list *list = malloc(sizeof(*list));
    list->used = 0;
    list->refs = 1;
    list->size = size;
    list->list = malloc(list->size*sizeof(list->list[0]));
    return list;
}

static void
sockcheck_list_append(struct sockcheck_list *list, struct sockcheck_state *new_item)
{
    if (list->used == list->size) {
	list->size <<= 1;
	list->list = realloc(list->list, list->size*sizeof(list->list[0]));
    }
    list->list[list->used++] = new_item;
}

static struct sockcheck_list *
sockcheck_list_clone(struct sockcheck_list *old_list)
{
    struct sockcheck_list *new_list = malloc(sizeof(*new_list));
    new_list->used = old_list->used;
    new_list->refs = 1;
    new_list->size = old_list->size;
    new_list->list = malloc(new_list->size*sizeof(new_list->list[0]));
    memcpy(new_list->list, old_list->list, new_list->used*sizeof(new_list->list[0]));
    return new_list;
}

static void
sockcheck_list_unref(struct sockcheck_list *list)
{
    if (!list || --list->refs > 0) return;
    free(list->list);
    free(list);
}

static void
sockcheck_issue_gline(sockcheck_cache_info sci)
{
    char *target = alloca(3+strlen(sci->hostname));
    strcpy(target, "*@");
    strcpy(target+2, sci->hostname);
    log(PC_LOG, LOG_INFO, "Issuing gline for client at IP %08x hostname %s: %s\n", sci->addr, sci->hostname, sci->reason);
    gline_add("ProxyCheck", target, 3600, sci->reason, 1);
}

void
sockcheck_queue_address(unsigned long addr)
{
    sockcheck_cache_info sci;
    unsigned char buff[17], *name;

    if (disable_sockcheck) return;
    if (pending_ip_pipe[1] == -1) return; /* something went wrong in worker */
    name = buff + 1;
    snprintf(name, sizeof(buff)-1, "%ld.%ld.%ld.%ld", (addr >> 0) & 255, (addr >> 8) & 255, (addr >> 16) & 255, (addr >> 24) & 255);
    if (pthread_mutex_lock(&checked_ip_mutex)) {
	/* lock failure; this usually means coding error */
	log(PC_LOG, LOG_ERROR, "BUG: pthread_mutex_lock(&checked_ip_mutex) returned errno %d (%s)\n", errno, strerror(errno));
	return;
    }
    sci = dict_find(checked_ip_dict, name, NULL);
    if ((sci != NULL)
	&& (sci->addr == addr)
	&& ((tv_now.tv_sec - sci->last_touched) < sockcheck_conf.max_cache_age)) {
	if (sci->decision == REJECT) sockcheck_issue_gline(sci);
    } else {
        int namelen = strlen(name) + 1;
	buff[0] = namelen;
	if (!sci) {
	    sci = malloc(sizeof(*sci) + namelen);
	    sci->decision = PENDING;
	    sci->last_touched = tv_now.tv_sec;
	    sci->reason = NULL;
	    sci->addr = addr;
	    memcpy(sci->hostname, name, namelen);
	    dict_insert(checked_ip_dict, sci->hostname, sci);
	}
	if (write(pending_ip_pipe[1], buff, namelen) < namelen) {
	    log(PC_LOG, LOG_ERROR, "Error: write(pending_ip_pipe[1]) returned errno %d (%s)\n", errno, strerror(errno));
	}
    }
    pthread_mutex_unlock(&checked_ip_mutex);
}

int
sockcheck_uncache_host(const char *name)
{
    int res;
    sockcheck_cache_info sci;

    if ((sci = dict_find(checked_ip_dict, name, NULL))) {
	if ((sci->decision == PENDING) || (sci->decision == CHECKING)) {
	    return -1;
	}
    }

    if (pthread_mutex_lock(&checked_ip_mutex)) {
	/* lock failure; this usually means coding error */
	log(PC_LOG, LOG_ERROR, "BUG: pthread_mutex_lock(&checked_ip_mutex) returned errno %d (%s)\n", errno, strerror(errno));
	return 0;
    }
    res = dict_remove(checked_ip_dict, name);
    pthread_mutex_unlock(&checked_ip_mutex);
    return res;
}

static unsigned int deq_buff_size, deq_buff_used;
static unsigned char *deq_buff;

static void
sockcheck_read_pending(void)
{
    if (deq_buff_used < deq_buff_size) {
	int nbr = read(pending_ip_pipe[0], deq_buff+deq_buff_used, deq_buff_size-deq_buff_used);
	if (nbr == 0) {
	    pthread_exit(NULL);
	} else if (nbr < 0) {
	    if (errno != EAGAIN) return;
	    nbr = 0;
	}
	deq_buff_used += nbr;
	if (deq_buff_used == deq_buff_size) {
	    deq_buff_size <<= 1;
	    deq_buff = realloc(deq_buff, deq_buff_size);
	}
    }
    while (deq_buff_used && deq_buff[0] && (deq_buff[0] <= deq_buff_used)) {
	unsigned char temp = deq_buff[deq_buff[0]];
	deq_buff[deq_buff[0]] = 0;
	string_list_append(pending_ip_list, strdup(deq_buff+1));
	deq_buff_used -= deq_buff[0];
	memmove(deq_buff, deq_buff+deq_buff[0], deq_buff_used);
	deq_buff[0] = temp;
    }
}

#if SOCKCHECK_DEBUG
static void
sockcheck_print_client(const struct sockcheck_client *client)
{
    static const char *decs[] = {"PENDING", "CHECKING", "ACCEPT", "REJECT"};
    log(PC_LOG, LOG_INFO, "client %p: { addr = %p { decision = %s; last_touched = " FMT_TIME_T "; reason = %s; addr = %08lx; hostname = \"%s\" }; "
        "test_index = %d; state = %p { port = %d; type = %s; template = \"%s\"; ... }; "
        "send = %p; send_size = %d; send_used = %d; send_pos = %d; "
        "read = %p; read_size = %d; read_used = %d; read_pos = %d; "
        "connected = %d; writing = %d; }\n",
        client, client->addr, decs[client->addr->decision], client->addr->last_touched, client->addr->reason, client->addr->addr, client->addr->hostname,
        client->test_index, client->state,
        client->state ? client->state->port : 0,
        client->state ? decs[client->state->type] : "N/A",
        client->state ? client->state->template : "N/A",
        client->send, client->send_size, client->send_used, client->send_pos,
        client->read, client->read_size, client->read_used, client->read_pos,
        client->connected, client->writing);
}
#endif

static unsigned char hexvals[256] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 16 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 32 */
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 0, 0, 0, 0, 0, /* 48 */
    0,10,11,12,13,14,15, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 64 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 80 */
    0,10,11,12,13,14,15, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 96 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 112 */

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 128 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 144 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 160 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 176 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 192 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 208 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 224 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 240 */
};

/* TODO: consider allowing non-NULL p_expansion with non-zero p_exp_length,
 * to avoid the malloc() impact.  If it's important for speed.  */
static void
expand_var(const struct sockcheck_client *client, unsigned char var, unsigned char **p_expansion, unsigned int *p_exp_length)
{
    extern struct cManagerNode cManager;
    const unsigned char *expansion;
    unsigned int exp_length;
#ifndef __SOLARIS__
    u_int32_t exp4;
    u_int16_t exp2;
#else
    uint32_t exp4;
    uint16_t exp2;
#endif
    /* expand variable */
    switch (var) {
    case 'c':
        expansion = client->addr->hostname;
        exp_length = strlen(expansion);
        break;
    case 'i':
	exp4 = client->addr->addr;
	exp_length = sizeof(exp4);
	expansion = (unsigned char*)&exp4;
	break;
    case 'p':
	exp2 = htons(client->state->port);
	exp_length = sizeof(exp2);
	expansion = (unsigned char*)&exp2;
	break;
    case 'u':
	expansion = cManager.uplink->host;
	exp_length = strlen(expansion);
	break;
    default:
	expansion = "";
	exp_length = 0;
    }
    if (p_expansion) {
	*p_expansion = malloc(exp_length);
	memcpy(*p_expansion, expansion, exp_length);
    }
    if (p_exp_length) {
	*p_exp_length = exp_length;
    }
}

static void
sockcheck_elaborate_state(struct pollfd *poller, struct sockcheck_client *client)
{
    const unsigned char *template = client->state->template;
    unsigned int n;

    if (template[0]) {
	client->writing = 1;
	client->send_used = 0;
	while (*template) {
	    unsigned char *target;
	    if (client->send_used == client->send_size) {
		client->send_size <<= 1;
		client->send = realloc(client->send, client->send_size);
	    }
	    target = client->send + client->send_used;
	    switch (*template) {
	    case '=': *target = template[1]; break;
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
	    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
	    case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
		*target = hexvals[template[0]] << 4 | hexvals[template[1]];
		break;
	    case '$': {
		unsigned char *expansion;
		unsigned int exp_length;
		expand_var(client, template[1], &expansion, &exp_length);
#if SOCKCHECK_DEBUG
                log(PC_LOG, LOG_INFO, "Expanded var %c for client %s to \"%s\" (len %d, pos %d)\n", template[1], client->addr->hostname, expansion, exp_length, client->send_used);
#endif
	        if (client->send_used+exp_length >= client->send_size) {
		    do {
			client->send_size <<= 1;
		    } while (client->send_used+exp_length >= client->send_size);
		    client->send = realloc(client->send, client->send_size);
		    target = client->send + client->send_used;
		}
		memcpy(target, expansion, exp_length);
		free(expansion);
		client->send_used += exp_length - 1;
	        break;
	    }
	    }
	    client->send_used++;
	    template += 2;
	}
	poller->events = POLLOUT;
    } else {
	client->writing = 0;
        client->send_used = 0;
	poller->events = POLLIN;
    }
    client->send_pos = 0;
    client->read_pos = 0;
    client->read_used = 0;
#if SOCKCHECK_DEBUG
    log(PC_LOG, LOG_INFO, "elaborated state: ");
    sockcheck_print_client(client);
#endif
    for (n=0; n<client->state->responses.used; n++) {
	client->resp_state[n] = client->state->responses.list[n]->template;
    }
}

static void
sockcheck_begin_test(unsigned int idx)
{
    struct sockcheck_client *client = client_list[idx];
    struct pollfd *poller = poll_list+idx+1;
    struct sockaddr_in sin;
    int flags;

    client->connected = 0;
    client->state = client->tests->list[client->test_index];
    if (!client->state) {
        client->test_index++;
        return;
    }

    client->read_timeout.tv_sec = tv_now.tv_sec + client->state->timeout / 1000;
    client->read_timeout.tv_usec = tv_now.tv_usec + (client->state->timeout % 1000) * 1000;
    if (client->read_timeout.tv_usec > 1000000) {
        client->read_timeout.tv_sec++;
        client->read_timeout.tv_usec -= 1000000;
    }

    poller->events = POLLIN | POLLOUT;
    poller->revents = 0; /* otherwise we get confused this poll iteration.. */
    if ((poller->fd = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
	log(PC_LOG, LOG_ERROR, "Error: socket() for client %s returned errno %d (%s)\n", client->addr->hostname, errno, strerror(errno));
	goto cleanup_1;
    }
    flags = fcntl(poller->fd, F_GETFL);
    fcntl(poller->fd, F_SETFL, flags|O_NONBLOCK);
    if (sockcheck_conf.local_addr) {
	if (bind(poller->fd, (struct sockaddr*) sockcheck_conf.local_addr, sizeof(struct sockaddr)) < 0) {
	    log(PC_LOG, LOG_ERROR, "Error: bind() for client %s returned errno %d (%s)\n", client->addr->hostname, errno, strerror(errno));
	}
    }
    sin.sin_family = AF_INET;
    sin.sin_port = htons(client->state->port);
    sin.sin_addr.s_addr = client->addr->addr;
    if (connect(poller->fd, (struct sockaddr*)&sin, sizeof(sin)) < 0) {
	if (errno != EINPROGRESS) {
	    log(PC_LOG, LOG_ERROR, "Error: connect() returned errno %d (%s) for port %d on client %s\n", errno, strerror(errno), client->state->port, client->addr->hostname);
	    /* progress to next test.. */
	    goto cleanup_2;
	}
    }
#if SOCKCHECK_DEBUG
    log(PC_LOG, LOG_INFO, "Starting proxy check on port %d (test %d) for client %s, fd %d\n", client->state->port, client->test_index, client->addr->hostname, poller->fd);
#endif
    return;

  cleanup_2:
    close(poller->fd);
  cleanup_1:
    poller->fd = -2;
    client->state = NULL;
    return;
}

static struct sockcheck_client *
sockcheck_alloc_client(sockcheck_cache_info sci)
{
    struct sockcheck_client *client = malloc(sizeof(*client));
    memset(client, 0, sizeof(*client));
    client->tests = tests;
    client->tests->refs++;
    client->addr = sci;
    client->send_size = 32;
    client->send = malloc(client->send_size);
    client->read_size = sockcheck_conf.max_read;
    client->read = malloc(client->read_size);
    client->resp_state = malloc(max_responses * sizeof(client->resp_state[0]));
    return client;
}

static void
sockcheck_free_client(struct sockcheck_client *client)
{
    sockcheck_list_unref(client->tests);
    free(client->send);
    free(client->read);
    free(client->resp_state);
    free(client);
}

static int
sockcheck_worker_queue_client(unsigned int idx)
{
    sockcheck_cache_info sci;
    const char *new_addr;

    if (pending_ip_list->used == 0) return 0;
    if (idx == (unsigned)-1) idx = sockcheck_num_sockets;
    if (idx == sockcheck_conf.max_sockets) return 0;
    new_addr = pending_ip_list->list[0];
    memmove(pending_ip_list->list, pending_ip_list->list+1,
	    (--pending_ip_list->used)*sizeof(pending_ip_list->list[0]));
    sockcheck_num_sockets++;
    if (pthread_mutex_lock(&checked_ip_mutex)) {
	/* lock failure; this usually means coding error */
	log(PC_LOG, LOG_ERROR, "BUG: pthread_mutex_lock(&checked_ip_mutex) returned errno %d (%s)\n", errno, strerror(errno));
	sci = NULL;
    } else {
	sci = dict_find(checked_ip_dict, new_addr, NULL);
	pthread_mutex_unlock(&checked_ip_mutex);
	if (!sci) {
	    log(PC_LOG, LOG_ERROR, "BUG: sockcheck_worker_queue_client(%d) could not find info for %s\n", idx, new_addr);
	    return 0;
	}
    }
    if (!tests) return 0;
    log(PC_LOG, LOG_INFO, "Proxy-checking client at %08x (%s) as client %d of %d\n", sci->addr, sci->hostname, idx, sockcheck_num_sockets);
    client_list[idx] = sockcheck_alloc_client(sci);
    client_list[idx]->test_rep = 0;
    sockcheck_begin_test(idx);
    return 1;
}

static void
sockcheck_decide(unsigned int idx, enum sockcheck_decision decision)
{
    struct sockcheck_client *client = client_list[idx];
    unsigned int n;

    checked_ip_count++;
    client->addr->decision = decision;
    client->addr->last_touched = tv_now.tv_sec;
    if (decision == ACCEPT) {
	/* do nothing */
#if SOCKCHECK_DEBUG
	log(PC_LOG, LOG_INFO, "Proxy check passed for client at IP %08x hostname %s\n", client->addr->addr, client->addr->hostname);
#endif
    } else if (decision == REJECT) {
	client->addr->reason = client->state->template;
	proxies_detected++;
	sockcheck_issue_gline(client->addr);
#if SOCKCHECK_DEBUG
	log(PC_LOG, LOG_INFO, "Proxy check failed for client at IP %08x hostname %s (%s)\n", client->addr->addr, client->addr->hostname, client->addr->reason);
#endif
	/* protect the tests list with a mutex.. 
	 * multiple threads shouldn't be rearranging the list
	 * -Jedi
	 */
	if (pthread_mutex_lock(&tests_mutex)) {
	    /* lock failure; this usually means coding error */
	    log(PC_LOG, LOG_ERROR, "BUG: pthread_mutex_lock(&tests_mutex) returned errno %d (%s)\n", errno, strerror(errno));
	    return;
	}

	/* Don't compare idx != 0 directly, because somebody else may have
	 * reordered the tests already. */
	if (client->tests->list[client->test_index] != tests->list[0]) {
	    struct sockcheck_list *new_tests = sockcheck_list_clone(tests);
	    struct sockcheck_state *new_first = client->tests->list[client->test_index];
            for (n=0; (n<tests->used) && (tests->list[n] != new_first); n++) ;
            for (; n>0; n--) new_tests->list[n] = new_tests->list[n-1];
	    new_tests->list[0] = new_first;
	    sockcheck_list_unref(tests);
	    tests = new_tests;
	}

	pthread_mutex_unlock(&tests_mutex);
    } else {
	log(PC_LOG, LOG_ERROR, "BUG: sockcheck_decide(\"%s\", %d): unrecognized decision\n", client->addr->hostname, decision);
    }
}

static void
sockcheck_try_write(int fd, struct sockcheck_client *client)
{
    /* Try to write remaining data from the send buffer. */
    int res = write(fd, client->send + client->send_pos, client->send_used - client->send_pos);
    if (res < 0) {
	if (errno == EAGAIN) return;
	log(PC_LOG, LOG_ERROR, "BUG: sockcheck_try_write(%s): write() returned errno %d (%s)\n", client->addr->hostname, errno, strerror(errno));
    } else {
	client->send_pos += res;
	if ((res == 0) || (client->send_pos == client->send_used)) {
	    client->writing = 0;
	}
    }
}

static int
sockcheck_try_read(int fd, struct sockcheck_client *client)
{
    /* read what we can from the fd */
    int res;
    unsigned int n;

    res = read(fd, client->read + client->read_used, client->read_size - client->read_used);
    if (res < 0) {
        if (errno == ECONNRESET) return -1;
	if (errno != EAGAIN) {
	    log(PC_LOG, LOG_ERROR, "BUG: sockcheck_try_read(%d, %s): read() returned errno %d (%s)\n", fd, client->addr->hostname, errno, strerror(errno));
	    return -1;
	}
    } else if (res == 0) {
	return client->state->responses.used - 1;
    } else {
	client->read_used += res;
    }

    /* Handle responses that dont require reading */
    if (client->read_pos == client->read_used) {
	for (n=0; n<(client->state->responses.used-1); n++) {
	    /* since we have nothing to read we will never enter the while() loop */
	    if (client->resp_state[n] && !strlen(client->resp_state[n])) {
#if SOCKCHECK_DEBUG
                log(PC_LOG, LOG_ERROR, "Choosing response %d for client %s\n", n, client->addr->hostname);
#endif
                return n;
	    }
	}
    }

    /* See if what's been read matches any of the expected responses */
    while (client->read_pos < client->read_used) {
	unsigned char curr, bleh;
	const unsigned char *resp_state;

	curr = client->read[client->read_pos++];
	for (n=0; n<(client->state->responses.used-1); n++) {
	    /* compare against possible target */
	    resp_state = client->resp_state[n];
	    if (resp_state == NULL) continue;
	    switch (*resp_state) {
	    case '=': if (curr != resp_state[1]) resp_state = NULL; break;
	    case '.': /* any character passes */ break;
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
	    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
	    case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
		bleh = hexvals[resp_state[0]] << 4 | hexvals[resp_state[1]];
		if (curr != (hexvals[resp_state[0]] << 4 | hexvals[resp_state[1]])) {
		    resp_state = NULL;
		}
		break;
	    case '$': {
		unsigned char *expansion;
		unsigned int exp_length;
		expand_var(client, resp_state[1], &expansion, &exp_length);
		if ((client->read_pos+exp_length-1) < client->read_used) {
		    if (memcmp(client->read+client->read_pos-1,
			       expansion, exp_length)) {
			resp_state = NULL;
		    } else {
			client->read_pos += exp_length - 1;
		    }
		} else {
		    /* can't check the variable yet, so come back later */
		    resp_state -= 2;
		}
		free(expansion);
		break;
	    }
	    }
	    if (resp_state) {
		client->resp_state[n] = resp_state = resp_state + 2;
		if (!*resp_state) {
		    log(PC_LOG, LOG_ERROR, "Choosing response %d for client %s\n", n, client->addr->hostname);
		    return n;
		}
	    } else {
		client->resp_state[n] = NULL;
	    }
	}
    }

    /* nothing seemed to match.  what now? */
    if (client->read_used >= client->read_size) {
	/* we got more data than we expected to get .. don't read any more */
#if SOCKCHECK_DEBUG
	log(PC_LOG, LOG_INFO, "Buffer filled (unmatched) for client %s\n", client->addr->hostname);
#endif
	return client->state->responses.used - 1;
    } else if (TV_CMP(tv_now, >, client->read_timeout)) {
	/* read timed out, use default handler. */
#if SOCKCHECK_DEBUG
	log(PC_LOG, LOG_INFO, "Read timeout expired for client %s\n", client->addr->hostname);
#endif
	return client->state->responses.used - 1;
    } else {
	/* keep waiting for more data. */
	return -1;
    }
}

/* Returns non-zero if this client is totally done */
static int
sockcheck_advance(unsigned int idx)
{
    struct sockcheck_client *client = client_list[idx];
    struct pollfd *poller = poll_list+idx+1;
    int next_state = -1;

    if (!client->state) {
        /* If we got here with a NULL state, connect() to the client
         * previously failed.  Advance it to the next test. */
        goto next_test;
    }

    if (!client->connected) {
	if (poller->revents & POLLHUP) {
	  disconn:
	    /* disconnected -- next test! */
	    close(poller->fd);
            goto next_test;
	} else if (poller->revents & (POLLOUT|POLLIN)) {
	    int rc, arglen = sizeof(rc);
	    if (getsockopt(poller->fd, SOL_SOCKET, SO_ERROR, &rc, &arglen) < 0) rc = errno;
	    switch (rc) {
	    default:
		log(PC_LOG, LOG_ERROR, "BUG: getsockopt(SO_ERROR) returned error %d (%s)\n", rc, strerror(rc));
	    case ECONNREFUSED:
	    case ETIMEDOUT:
#if SOCKCHECK_DEBUG
                log(PC_LOG, LOG_INFO, "Client %s gave us errno %d (%s)\n", client->addr->hostname, rc, strerror(rc));
#endif
		goto disconn;
	    case 0: break;
	    }
	    client->connected = 1;
	    sockcheck_elaborate_state(poller, client);
	} else if (TV_CMP(tv_now, >, client->read_timeout)) {
	    goto disconn;
	}
	return 0;
    }

    if (client->writing) {
	if ((poller->revents & POLLHUP)
	    || TV_CMP(tv_now, >, client->read_timeout)) {
	    /* hung up, fall back to default for this state */
	    next_state = (signed)client->state->responses.used - 1;
	} else {
	    sockcheck_try_write(poller->fd, client);
	}
	if (!client->writing) {
	    poller->events = POLLIN;
	}
    } else {
	if (poller->revents & POLLHUP) {
	    next_state = (signed)client->state->responses.used - 1;
	} else {
	    next_state = sockcheck_try_read(poller->fd, client);
	}
	if (client->writing) {
	    poller->events = POLLOUT;
	}
    }

    if (next_state != -1) {
	struct sockcheck_state *ns = client->state->responses.list[next_state]->next;
#if SOCKCHECK_DEBUG
	/* debug code, not for general use */
	unsigned int n, m;
	char buffer[201];
	static const char *hexmap = "0123456789ABCDEF";
	log(PC_LOG, LOG_INFO, "sockcheck_advance(%s) following response %d (type %d) of %d\n", client->addr->hostname, next_state+1, ns->type, client->state->responses.used);
	for (n=0; n<client->send_used; n++) {
	    for (m=0; (m<(sizeof(buffer)-1)>>1) && ((n+m) < client->send_used); m++) {
		buffer[m << 1] = hexmap[client->send[n+m] >> 4];
		buffer[m << 1 | 1] = hexmap[client->send[n+m] & 15];
	    }
	    buffer[m<<1] = 0;
	    log(PC_LOG, LOG_INFO, "written data: %s\n", buffer);
	    n += m;
	}
	for (n=0; n<client->read_used; n++) {
	    for (m=0; (m<(sizeof(buffer)-1)>>1) && ((n+m) < client->read_used); m++) {
		buffer[m << 1] = hexmap[client->read[n+m] >> 4];
		buffer[m << 1 | 1] = hexmap[client->read[n+m] & 15];
	    }
	    buffer[m<<1] = 0;
	    log(PC_LOG, LOG_INFO, "read data: %s\n", buffer);
	    n += m;
	}
	log(PC_LOG, LOG_INFO, "advancing client: ");
        sockcheck_print_client(client);
#endif
	client->state = ns;
	switch (ns->type) {
	case CHECKING: break;
	case REJECT:
	    close(poller->fd);
	    sockcheck_decide(idx, REJECT);
            return 1;
	case ACCEPT:
	    close(poller->fd);
          next_test:
            if (++client->test_rep < client->tests->list[client->test_index]->reps) {
                sockcheck_begin_test(idx);
            } else if (++client->test_index < client->tests->used) {
                client->test_rep = 0;
                sockcheck_begin_test(idx);
            } else {
		sockcheck_decide(idx, ACCEPT);
                return 1;
	    }
	    break;
	default:
	    log(PC_LOG, LOG_ERROR, "BUG: unknown next-state type %d (after %p)\n", ns->type, client->state);
	    break;
	}
    }
    return 0;
}

static int
find_old_entries(const char *hostname, void *data, void *extra)
{
    sockcheck_cache_info sci = data;
    struct sci_list *cleaned = extra;

    (void)hostname;
    switch (sci->decision) {
    case ACCEPT:
    case REJECT:
        /* Forget anything more than an hour old. */
        if (sci->last_touched < (tv_now.tv_sec - 3600)) sci_list_append(cleaned, sci);
        break;
    default:
        /* Do nothing. */
        break;
    }
    return 0;
}

static void *
sockcheck_worker(void *arg)
{
    int res, timeout, err_count=0;
    unsigned int n;
    struct sci_list cleaned;

    (void)arg;
    sci_list_init(&cleaned);
    while (1) {
	timeout = sockcheck_num_sockets ? 1000 : -1;
	res = poll(poll_list, sockcheck_num_sockets+1, timeout);
	if (pending_ip_pipe[0] == -1) break;
	if (res >= 0) {
	    if (gettimeofday(&tv_now, NULL) < 0) {
		log(PC_LOG, LOG_ERROR, "BUG: sockcheck_worker(): gettimeofday() returned errno %d (%s)\n", errno, strerror(errno));
	    }
            /* This isn't done with timeq because timeq is in the wrong thread :( */
            if (tv_now.tv_sec > (last_clean + 3600)) {
                cleaned.used = 0;
                dict_foreach(checked_ip_dict, find_old_entries, &cleaned);
                for (n=0; n<cleaned.used; n++) {
                    dict_remove(checked_ip_dict, cleaned.list[n]->hostname);
                }
		last_clean = tv_now.tv_sec;
            }
            /* Do we have any new requests or connetions? */
	    err_count = 0;
	    if (poll_list[0].revents) sockcheck_read_pending();
            while (sockcheck_worker_queue_client(-1)) ;
            /* Handle existing connections */
	    for (n=0; n<sockcheck_num_sockets; n++) {
                /* Do work.  If the client's not done, continue. */
		if (!sockcheck_advance(n)) continue;
                /* Testing done.  Free client and decrement number of sockets in progress. */
                sockcheck_num_sockets--;
                sockcheck_free_client(client_list[n]);
                /* Swap down the highest-numbered socket client, if necessary. */
                if (n != sockcheck_num_sockets) {
#if SOCKCHECK_DEBUG
                    log(PC_LOG, LOG_INFO, "sockcheck moving client %s (at %d) to fill hole at %d\n", client_list[sockcheck_num_sockets]->addr->hostname, sockcheck_num_sockets, n);
#endif
                    poll_list[n+1] = poll_list[sockcheck_num_sockets+1];
                    client_list[n] = client_list[sockcheck_num_sockets];
                }
                client_list[sockcheck_num_sockets] = 0;
                n--;
	    }
	} else if (res == -1) {
	    if (errno == EINTR) continue;
	    log(PC_LOG, LOG_ERROR, "Error: poll(poll_list,...) returned errno %d (%s)\n", errno, strerror(errno));
	    if (++err_count > 4) {
		log(PC_LOG, LOG_ERROR, "Too many errors in worker thread, killing it\n");
 		close(pending_ip_pipe[0]);
 		close(pending_ip_pipe[1]);
		pending_ip_pipe[1] = -1;
		break;
	    }
	}
    }
    sci_list_clean(&cleaned);
    return NULL;
}

static int
sockcheck_create_response(const char *key, void *data, void *extra)
{
    const char *str, *end;
    struct record_data *rd = data;
    struct sockcheck_state *parent = extra;
    struct sockcheck_response *resp;
    dict_t resps;
    char *templ;

    /* allocate memory and tack it onto parent->responses */
    resp = malloc(sizeof(*resp));
    for (end = key; *end != ':' && *end != 0; end += 2 && end) ;
    templ = malloc(end - key + 1);
    memcpy(templ, key, end - key);
    templ[end - key] = 0;
    resp->template = templ;
    resp->next = malloc(sizeof(*resp->next));
    resp->next->port = parent->port;
    response_list_append(&parent->responses, resp);
    /* now figure out how to create resp->next */
    if ((str = GET_RECORD_QSTRING(rd))) {
	if (!strncasecmp(str, "reject", 6)) {
	    resp->next->type = REJECT;
	} else if (!strncasecmp(str, "accept", 6)) {
	    resp->next->type = ACCEPT;
	} else {
	    log(PC_LOG, LOG_ERROR, "Error: unknown sockcheck decision `%s', defaulting to accept\n", str);
	    resp->next->type = ACCEPT;
	}
	if (str[6]) {
	    resp->next->template = strdup(str+7);
	} else {
	    resp->next->template = strdup("No explanation given");
	}
    } else if ((resps = GET_RECORD_OBJECT(rd))) {
	resp->next->type = CHECKING;
	response_list_init(&resp->next->responses);
	if (*end == ':') {
	    resp->next->template = strdup(end+1);
	} else {
	    resp->next->template = strdup("");
	}
	dict_foreach(resps, sockcheck_create_response, resp->next);
    }
    return 0;
}

/* key: PORT:send-pattern, as in keys of sockcheck.db.example
 * data: recdb record_data containing response
 * extra: struct sockcheck_list* to append test to
 */
static int
sockcheck_create_test(const char *key, void *data, void *extra)
{
    char *end;
    struct record_data *rd;
    dict_t object;
    struct sockcheck_state *new_test;
    unsigned int n;

    rd = data;
    new_test = malloc(sizeof(*new_test));
    new_test->template = NULL;
    new_test->reps = 1;
    new_test->port = strtoul(key, &end, 0);
    new_test->timeout = 5000;
    new_test->type = CHECKING;
    response_list_init(&new_test->responses);
    if (!(object = GET_RECORD_OBJECT(rd))) {
	log(PC_LOG, LOG_ERROR, "Error: misformed sockcheck test `%s', skipping it\n", key);
        free(new_test);
        return 1;
    }
    while (*end) {
        switch (*end) {
        case '*': new_test->reps = strtoul(end+1, &end, 0); break;
        case ':': new_test->template = strdup(end+1); end += strlen(end); break;
        default:
            log(PC_LOG, LOG_ERROR, "Error: misformed sockcheck test `%s', skipping it\n", key);
            free(new_test);
            return 1;
        }
    }
    if (!new_test->template) {
	log(PC_LOG, LOG_ERROR, "Error: misformed sockcheck test `%s', skipping it\n", key);
	free(new_test);
	return 1;
    }
    dict_foreach(object, sockcheck_create_response, new_test);
    /* If none of the responses have template "other", create a
     * default response that goes to accept. */
    for (n=0; n<new_test->responses.used; n++) {
	if (!strcmp(new_test->responses.list[n]->template, "other")) break;
    }
    if (n == new_test->responses.used) {
	rd = alloc_record_data_qstring("accept");
	sockcheck_create_response("other", rd, new_test);
	free_record_data(rd);
    } else if (n != (new_test->responses.used - 1)) {
	struct sockcheck_response *tmp;
	/* switch the response for "other" to the end */
	tmp = new_test->responses.list[new_test->responses.used - 1];
	new_test->responses.list[new_test->responses.used - 1] = new_test->responses.list[n];
	new_test->responses.list[n] = tmp;
	
    }
    if (new_test->responses.used > max_responses) {
	max_responses = new_test->responses.used;
    }
    sockcheck_list_append(extra, new_test);
    return 0;
}

static void
sockcheck_read_tests(void)
{
    dict_t test_db;
    struct sockcheck_list *new_tests;
    test_db = parse_database(SOCKCHECK_TEST_DB);
    if (!test_db) {
	disable_sockcheck = 1;
	return;
    }
    if (dict_size(test_db) > 0) {
	new_tests = sockcheck_list_alloc(dict_size(test_db));
	dict_foreach(test_db, sockcheck_create_test, new_tests);
	if (tests) sockcheck_list_unref(tests);
	tests = new_tests;
    } else {
	log(PC_LOG, LOG_ERROR, "%s was empty - disabling sockcheck\n", SOCKCHECK_TEST_DB);
	disable_sockcheck = 1;
    }
    free_database(test_db);
}

static void _sockcheck_init(void);
static void sockcheck_shutdown(void);

void
sockcheck_read_conf(void)
{
    struct record_data *rd;
    dict_t my_node;
    const char *str;
    /* Keep a copy of disable_sockcheck so we know if it got changed. */
    int disable = disable_sockcheck;

    rd = conf_get_node("sockcheck");
    /* set the defaults here in case the entire record is missing */
    sockcheck_conf.max_sockets = 32;
    sockcheck_conf.max_read = 1024;
    sockcheck_conf.max_cache_age = 24*60*60; /* 1 day */
    if (sockcheck_conf.local_addr) {
        free(sockcheck_conf.local_addr);
        sockcheck_conf.local_addr = NULL;
    }
    /* now try to read from the conf database */
    if (rd && (my_node = GET_RECORD_OBJECT(rd))) {
	str = database_get_data(my_node, "enabled", RECDB_QSTRING);
	if (str) disable_sockcheck = !atoi(str);

	str = database_get_data(my_node, "max_sockets", RECDB_QSTRING);
	if (str) sockcheck_conf.max_sockets = strtoul(str, NULL, 0);
	str = database_get_data(my_node, "max_read", RECDB_QSTRING);
	if (str) sockcheck_conf.max_read = strtoul(str, NULL, 0);
	str = database_get_data(my_node, "max_cache_age", RECDB_QSTRING);
	if (str) sockcheck_conf.max_cache_age = ParseInterval(str);
	str = database_get_data(my_node, "address", RECDB_QSTRING);
	if (str) {
	    struct sockaddr_in *sin;
	    unsigned long addr;

	    sockcheck_conf.local_addr_len = sizeof(*sin);
	    if (getipbyname(str, &addr)) {
		sin = malloc(sockcheck_conf.local_addr_len);
		sin->sin_family = AF_INET;
		sin->sin_port = 0;
		sin->sin_addr.s_addr = addr;
#ifdef HAVE_SIN_LEN
	        sin->sin_len = 0;
#endif
        	memset(sin->sin_zero, 0, sizeof(sin->sin_zero));
	        sockcheck_conf.local_addr = sin;
	    } else {
		log(PC_LOG, LOG_ERROR, "Error: Unable to get host named `%s', not checking a specific address\n", str); 
		sockcheck_conf.local_addr = NULL;
	    }
	}
    }

    if (disable != disable_sockcheck) {
	/* The setting got changed. */
	if (!disable_sockcheck) {
	    /* Shut down, then reinitialize sockcheck. Disabling
	       sockcheck doesn't require any further action.*/
	    sockcheck_shutdown();
	    _sockcheck_init();
	}
    }
}

void
sockcheck_free_state(struct sockcheck_state *state)
{
    unsigned int n;
    if (state->type == CHECKING) {
	for (n=0; n<state->responses.used; n++) {
	    free((char*)state->responses.list[n]->template);
	    sockcheck_free_state(state->responses.list[n]->next);
	    free(state->responses.list[n]);
	}
	response_list_clean(&state->responses);
    }
    free((char*)state->template);
    free(state);
}

const char *
sockcheck_add_test(const char *desc)
{
    struct sockcheck_list *new_tests;
    const char *reason;
    char *name;
    struct record_data *rd;

    if ((reason = parse_record_string(desc, &name, &rd))) return reason;
    new_tests = sockcheck_list_clone(tests);
    if (sockcheck_create_test(name, rd, new_tests)) {
	sockcheck_list_unref(new_tests);
	return "Sockcheck test parse error";
    }
    sockcheck_list_unref(tests);
    tests = new_tests;
    return 0;
}

static void
sockcheck_shutdown(void)
{
    unsigned int n;
    char whut = 0;
    n = pending_ip_pipe[1];
    pending_ip_pipe[0] = pending_ip_pipe[1] = -1;
    write(n, &whut, 1);
    pthread_join(sockcheck_thread, NULL);
    dict_delete(checked_ip_dict);
    free(pending_ip_list); /* not free_string_list because it doesn't own them */
    free(deq_buff);
    if (tests) {
	for (n=0; n<tests->used; n++) {
	    sockcheck_free_state(tests->list[n]);
	}
    }
    sockcheck_list_unref(tests);
    free(poll_list);
    sockcheck_num_sockets = 0;
    if (sockcheck_conf.local_addr) {
	free(sockcheck_conf.local_addr);
	sockcheck_conf.local_addr_len = 0;
    }

    if (disable_sockcheck) return;

    if (client_list) {
	for (n=0; n<sockcheck_conf.max_sockets; n++) {
            if (client_list[n]) {
                sockcheck_free_client(client_list[n]);
            }
	}
	free(client_list);
    }
}

static void
_sockcheck_init(void)
{
    int flags;

    checked_ip_dict = dict_new();
    dict_set_free_data(checked_ip_dict, free);
    pending_ip_list = alloc_string_list(16);
    deq_buff_used = 0;
    deq_buff_size = 64;
    deq_buff = malloc(deq_buff_size);
    sockcheck_num_sockets = 0;

    /* poll_list[0] is used for pending_ip_pipe (see below) */
    poll_list = malloc((sockcheck_conf.max_sockets+1) * sizeof(poll_list[0]));
    if (pipe(pending_ip_pipe)) {
	log(PC_LOG, LOG_ERROR, "pipe(pending_ip_pipe) returned errno %d (%s)\n", errno, strerror(errno));
	return;
    }
    flags = fcntl(pending_ip_pipe[0], F_GETFL);
    fcntl(pending_ip_pipe[0], F_SETFL, flags|O_NONBLOCK);
    poll_list[0].fd = pending_ip_pipe[0];
    poll_list[0].events = POLLIN;
    sockcheck_read_tests();
    if (disable_sockcheck) return;
    client_list = calloc(sockcheck_conf.max_sockets, sizeof(client_list[0]));
    last_clean = now;
    if (pthread_create(&sockcheck_thread, NULL, sockcheck_worker, NULL)) {
	log(PC_LOG, LOG_ERROR, "pthread_create(..., sockcheck_worker, ...) returned errno %d (%s)\n", errno, strerror(errno));
	return;
    }
}

void
sockcheck_init(void)
{
    conf_register_reload(sockcheck_read_conf);
    sockcheck_read_conf();
    reg_exit_func(sockcheck_shutdown);

    _sockcheck_init();
}

void
sockcheck_get_stats(struct sockcheck_stats *stats)
{
    stats->checked_ip_count = checked_ip_count;
    stats->failed_ip_count = proxies_detected;
    stats->checking_ip_count = sockcheck_num_sockets;
    stats->checking_ip_max = sockcheck_conf.max_sockets;
    stats->pending_ip_count = pending_ip_list->used;
    stats->test_count = tests ? tests->used : 0;
    stats->cached_ip_count = dict_size(checked_ip_dict);
}

#else

void
sockcheck_init(void)
{
#if !defined(HAVE_POLL)
    disable_sockcheck = 1;
    log(MAIN_LOG, LOG_INFO, "WARNING: no sockcheck support without poll(2).\n");
#elif !defined(HAVE_PTHREAD)
    disable_sockcheck = 1;
    log(MAIN_LOG, LOG_INFO, "WARNING: no sockcheck support without a thread library.\n");
#elif defined(__CYGWIN__)
    disable_sockcheck = 1;
    log(MAIN_LOG, LOG_INFO, "WARNING: no sockcheck under cygwin.\n");
#endif
}

#endif
